# (C) Michael V Ivanov <mike@if-site.com>
# All rights reserved
# this software is distributed under the same license as Zope itself

import PIL.Image
import wstring
import cStringIO,os

from freetype import *

class _CacheItem :
    def __init__( self,g,m ) :
	self.glyph = g
	self.metric = m

def _line_width ( cache,s ) :
    w = 0
    for char in s :
	w = w + cache[ char ].metric.advance
    return w / 64

def render_text (self,REQUEST=None) :

    font    = self.font_face
    size    = self.font_size
    font_en = self.font_encoding
    res_x   = self.resolution_x
    res_y   = self.resolution_y

    text    = self.text
    h_align = self.horizontal_align
    v_align = self.vertical_align
    wrap    = self.wrap
    charset = self.text_encoding

    fix_w    = self.fixed_width
    fix_h    = self.fixed_height
    bg_color = string.atoi( self.background_color, 16)
    fg_color = string.atoi( self.foreground_color, 16)
    format   = self.format

    if charset != "UNICODE" :
	__import__( os.path.join("wstring",charset))
    if font_en != "UNICODE" :
	__import__( os.path.join("wstring",font_en))

    bg_color = \
	((bg_color & 0x0000ff) << 16) |\
	((bg_color & 0x00ff00)) | \
	((bg_color & 0xff0000) >> 16)

    fg_color = \
	((fg_color & 0x0000ff) << 16) |\
	((fg_color & 0x00ff00)) | \
	((fg_color & 0xff0000) >> 16)

    # INITALIZE ENGINE

    engine = TT_Init_FreeType()
    TT_Set_Raster_Gray_Palette( engine,(0, 64, 128, 192, 255) )

    # PREPARE FONT

    font = string.replace( font, '/','')
    font = string.replace( font, '\\','')
    font = font + ".ttf"
    fontp = os.path.join( INSTANCE_HOME,'var','fonts',font )
    face = TT_Open_Face( engine, fontp )

    if face is None :
	raise "Resource not found", "Couldn't open font '%s'" % fontp
    
    props  = TT_Get_Face_Properties( face )

    # FIND CHARACTER MAP

    charmap = None
    for i in range(props.num_CharMaps ) :
	tup = TT_Get_CharMap_ID( face,i )
	if tup == (3,1) :
            # Windows Unicode mapping
            charmap = TT_Get_CharMap( face,i )
            break

    if not charmap :
	raise "No appropriate character map found!"

    # CALCULATE INSTANCE PARAMETERS

    inst   = TT_New_Instance( face )

    TT_Set_Instance_Resolutions( inst,res_x,res_y )
    TT_Set_Instance_PointSize( inst,size )
    metrics= TT_Get_Instance_Metrics( inst )

    scale_x = metrics.x_scale / float(0xffff)
    scale_y = metrics.y_scale / float(0xffff)

    header = props.header
    min_x = int( header.xMin * scale_x / 64)
    min_y = int( header.yMin * scale_y / 64)
    glyph_w = int(( header.xMax - header.xMin) * scale_x / 64)
    glyph_h = int(( header.yMax - header.yMin) * scale_y / 64)

    hh = props.horizontal
    ascend   = int( hh.Ascender * scale_y / 64)
    descend  = int( hh.Descender * scale_y / 64)
    line_gap = int( hh.Line_Gap * scale_y / 64)
    line_h = int((hh.Ascender - hh.Descender + hh.Line_Gap) * scale_y / 64)

    # PREPARE SYMBOL CACHE

    text = string.replace( text, "\r", "")

    cache = {}
    for char in text+' ' :
	if font_en == "UNICODE" :
	    idx = wstring.ord( wstring.decode( charset,char ))
	else :
	    if font_en == charset :
		idx = ord( char)
	    else :
		idx = ord( wstring.decode( charset,char ).encode( font_en))

	try :
	    idx = TT_Char_Index(charmap,idx)
	except :
	    idx = 0

	if not cache.has_key( char ) :
            glyph = TT_New_Glyph( face)
            TT_Load_Glyph( inst,glyph,idx,TTLOAD_DEFAULT )
            metric = TT_Get_Glyph_Metrics( glyph)
	    cache[ char ] = _CacheItem( glyph,metric )

    # CREATE RASTER MAP AND FG PATTERN

    raster = TT_Raster_Map( glyph_w,glyph_h,"pix" )
    fg_pat = PIL.Image.new("RGB",(raster.cols,raster.rows),fg_color)

    # PREPARE LINES

    if fix_w <= 0 :
	width = 0
	if wrap == "NONE" :
	    t = string.replace( text, "\n", " ")
	    lines = (t,)
	else :
	    wrap = "AS_IS"
	    lines = string.split( text, "\n")

	for line in lines :
	    width = max( width, _line_width( cache,line))
	    
    else :
	width = fix_w
	if wrap == "NONE" :
	    t = string.replace( text, "\n", " ")
	    lines = (t,)
	elif wrap == "AS_IS" :
	    lines = string.split( text, "\n")
	elif wrap == "WORD" :
	    words = string.split( text)
	    lines = []
	    line = ""
	    for word in words :
		if line == "" :
		    w = word
		else :
		    w = " " + word
		if _line_width( cache,line + w) <= fix_w :
		    line = line + w
		else :
		    if line == "" :
			line = line + word
			lines.append( line)
			line = ""
		    else :
			lines.append( line)
			line = word

	    if line != "" : lines.append( line)

	else :
	    t = string.replace( text, "\n", "")
	    lines = []
	    line = ""
	    for char in t :
		if _line_width( cache,line+char) <= fix_w :
		    line = line + char
		else :
		    lines.append( line)
		    line = char

	    if line != "" : lines.append( line)

    if fix_h <= 0 :
	height = line_h * len(lines)
	v_align = "TOP"
    else :
	height = fix_h

    # PREPARE RESULT IMAGE

    image = PIL.Image.new("RGB",(width,height),bg_color)

    # DRAW GLYPHS

    if v_align == "TOP" :
	from_y = 0
    elif v_align == "CENTER" :
	from_y = (height / 2) - (line_h * len(lines) / 2)
    else :
	from_y = height - (line_h * len(lines))

    cur_y = from_y

    for line in lines :
	if h_align == "LEFT" :
	    from_x = 0
	elif h_align == "CENTER" :
	    from_x = (width / 2) - (_line_width( cache,line ) / 2)
	else :
	    from_x = width - _line_width( cache,line )

	cur_x = from_x

	for char in line :
	    m = cache[ char ].metric
	    TT_Get_Glyph_Pixmap(
		cache[ char ].glyph, raster,
		-min_x*64, -min_y*64
		)
	    mask = PIL.Image.fromstring(
		"L",
		(raster.cols,raster.rows),
		raster.tostring(),
		"raw","L"
		)

	    memset( raster.bitmap, chr(0), raster.size)
	    #image.paste( fg_pat, (cur_x-min_x+m.bearingX/64, cur_y+(line_h-m.bearingY/64)),mask)
	    image.paste( fg_pat, 
		(
		    cur_x+min_x,
		    cur_y+min_y
		),
		mask
		)
	    cur_x = cur_x + cache[ char ].metric.advance / 64

	cur_y = cur_y + line_h

    out = cStringIO.StringIO()
    image.save( out, format)
    out.seek(0)
    self.update_data( out.read(), content_type="image/" + format)
    

